@@ -20,10 +20,28 @@
 ;; along with this program; if not, write to the Free Software
 ;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 
+;; speak: supported
+;; lexicon: ignored (no reasonable specification of the element)
+;; meta: ignored (no reasonable use here known)
+;; metadata: ignored (no reasonable use here known)
+;; p: supported
+;; s: supported
+;; say-as: trivially rendered (attribute values undefined in the standard)
+;; phoneme: unsupported
+;; sub: supported
+;; voice: supported, but prosody not yet retained (TODO)
+;; emphasis: not properly supported (TODO)
+;; break: partially supported (TODO)
+;; prosody: partially supported (TODO)
+;; audio: supported (TODO more complicated alternative text may fail to render)
+;; mark: supported
+;; desc: ignored (no reasonable use here known)
+
 
 (require_module 'rxp)
 
 (require 'prosody-param)
+(require 'recode)
 (require 'tokenize)
 (require 'util)
 (require 'voice-select)
@@ -80,7 +98,9 @@
 
 
 (define (ssml-attval attlist att)
-  (car (xxml_attval att attlist)))  
+  (let ((attval (car (xxml_attval att attlist))))
+    (if attval
+        (recode-utf8->current attval))))
 
 (define (ssml-attval-time attlist att)
   (let ((time (string-before (ssml-attval attlist 'time) "s")))
@@ -119,10 +139,12 @@
 (define (ssml-change-voice lang-code gender age variant name)
   (push current-voice ssml-voices)
   (push voice-select-current-defaults ssml-voice-parameters)
-  (select-voice* lang-code gender age variant name))
+  (select-voice* lang-code gender age variant name)
+  (restore-prosody))
 
 (define (ssml-unchange-voice)
   (voice.select (pop ssml-voices))
+  (restore-prosody)
   (set! voice-select-current-defaults (pop ssml-voice-parameters)))
 
 (define (ssml-change-language attlist)
@@ -140,7 +162,8 @@
     (while prosody
       (let ((id (first (car prosody)))
             (value (second (car prosody))))
-        (push (list id ((avalue-get id ssml-prosody-setters) value))
+        (push (list id (change-prosody (avalue-get id ssml-prosody-setters)
+                                       value))
               old-prosody))
       (set! prosody (cdr prosody)))
     old-prosody))
@@ -168,8 +191,15 @@
   (format t "Mark reached: %s\n" mark))
 
 (define (ssml-utt-text utt)
-  (let ((last-token (and utt (utt.relation.last utt 'Token))))
-    (and last-token (item.name last-token))))
+  (let* ((last-token (and utt (utt.relation.last utt 'Token)))
+         (token last-token)
+         (token-list nil))
+    (when last-token
+      (while (and token (not (eq? (item.feat token 'ssml-tag) 'noticed)))
+        (push (item.name token) token-list)
+        (set! token (item.prev token)))
+      (item.set_feat last-token 'ssml-tag 'noticed))
+    (apply string-append token-list)))
 
 
 ;;; Markup handlers
@@ -327,8 +357,6 @@
 
 (define ssml-audio-uri nil)
 (define (ssml.audio.start attlist utt)
-  ;; TODO: Ensure the alternative text is not spoken because of
-  ;; ssml-process-utt call inside it.
   (ssml-process-utt utt)
   (set! ssml-audio-uri (ssml-attval attlist 'src))
   nil)
@@ -419,7 +447,9 @@
   (set! ssml-xxml-elements.orig xxml_elements)
   (set! xxml_elements ssml-elements)
   (set! ssml-voices (list current-voice))
-  (reset-voice))
+  ;; reset-voice used to be called here, but it's not much useful and it resets
+  ;; current speechd settings, which is not what we want.
+  )
 
 (define (ssml_exit_func)
   (voice.select (car (last ssml-voices)))
@@ -447,6 +477,10 @@
       (fclose fd))
     (tts_file ssml-file 'ssml)))
 
+(define (ssml-say* ssml-text)
+  (ssml-parse ssml-text)
+  (ssml-speak-chunks))
+
 (define (ssml-parse ssml-text)
   (set! ssml-parsed '())
   (set! ssml-utterances '())
@@ -484,6 +518,8 @@
           (ssml-next-chunk))))
    (ssml-utterances
     (let ((utt (pop ssml-utterances)))
+      (if (eq? (typeof utt) 'string)
+          (set! utt (intern utt)))
       (if (or (symbol? utt) (utt.relation.items utt 'Token))
           utt
           (ssml-next-chunk))))
@@ -505,22 +541,40 @@
               (ssml-process-mark
                (lambda (mark)
                  (set! ssml-utterances (append ssml-utterances (list mark))))))
+        ;; Texts of some marks should be processed in a single step
+        (cond
+         ((member func-name '(sub.start prosody.start audio.start))
+          (push t ssml-in-volatile))
+         ((member func-name '(sub.end prosody.end audio.end))
+          (pop ssml-in-volatile)))
+        ;; If there's a text, recode it and add it to the utterance
         (if text
-            (begin
-              (if (or ssml-in-volatile
-                      (eq? func-name 'ssml.break))
+            (let ((lang-change (ssml-attval attlist 'xml:lang)))
+              (if lang-change
+                  (ssml-change-language attlist))
+              (set! text (recode-utf8->current text))
+              (if lang-change
+                  (ssml-unchange-language))
+              (if (or ssml-in-volatile (eq? func-name 'ssml.break))
                   (while (not (equal? text ""))
                     (set! text (get-token utt text)))
                   (set! ssml-current-text text))
               (push (list func-name attlist nil) ssml-parsed))
-            (begin
-              (cond
-               ((member func-name '(sub.start prosody.start audio.start))
-                (push t ssml-in-volatile))
-               ((member func-name '(sub.end prosody.end audio.end))
-                (pop ssml-in-volatile)))
-              (set! ssml-current-utt ((symbol-value func-name) attlist utt))))
-        (ssml-next-chunk))))))
+            ;; If there's no text, just call the mark function
+            (set! ssml-current-utt ((symbol-value func-name) attlist utt))))
+      (ssml-next-chunk)))))
 
+(define (ssml-speak-chunks)
+  (let ((utt (ssml-next-chunk)))
+    (cond
+     ((not utt)
+      nil)
+     ((symbol? utt)
+      (print utt)
+      (ssml-speak-chunks))
+     (t
+      (utt.synth utt)
+      (event-eat-utt utt wave.play)
+      (ssml-speak-chunks)))))
 
 (provide 'ssml-mode)
